#include "Chunk.hpp"
#include "AnvilChunk.hpp"
#include "nbt.hpp"
#include "Convert.hpp"

Section::Section()
{
	y = -1;

	blocks = new BYTE[SECTIONBLOCKS];
	blockData = new BYTE[SECTIONBLOCKS / 2];
	skyLight = new BYTE[SECTIONBLOCKS / 2];
	blockLight = new BYTE[SECTIONBLOCKS / 2];

	memset(blocks, 0, sizeof(BYTE) * SECTIONBLOCKS);
	memset(blockData, 0, sizeof(BYTE) * SECTIONBLOCKS / 2);
	memset(skyLight, 0, sizeof(BYTE) * SECTIONBLOCKS / 2);
	memset(blockLight, 0, sizeof(BYTE) * SECTIONBLOCKS / 2);
}

Section::~Section()
{
	delete [] blocks;
	delete [] blockData;
	delete [] skyLight;
	delete [] blockLight;
}

AnvilChunk::AnvilChunk() : Chunk()
{
	anvilHeightMap = 0;
	sections = 0;
}

AnvilChunk::~AnvilChunk()
{
}

void AnvilChunk::Allocate()
{
	if (!anvilHeightMap)
		anvilHeightMap = new int[MAPX * MAPY];
	if(!sections)
		sections = new Section*[MAXSECTIONS];

	memset(anvilHeightMap, 0, sizeof(int) * MAPX * MAPY);
	memset(sections, 0, sizeof(Section*) * MAXSECTIONS);
}

void AnvilChunk::Deallocate()
{
	if (anvilHeightMap)
	{
		delete[] anvilHeightMap;
		anvilHeightMap = 0;
	}

	if(sections)
	{
		for(int i = 0; i < MAXSECTIONS; i++)
		{
			if(sections[i])
			{
				delete sections[i];
				sections[i] = 0;
			}
		}

		delete[] sections;
		sections = 0;
	}
}

bool AnvilChunk::ParseNBT(void * p)
{
	if (!p)
		return false;
	NBT_Reader * read = (NBT_Reader *)p;

	Section* section = 0;
	int count = 0;
	NBTTag tag = read->ReadData(); // Read first compound
	while ((tag = read->ReadData()) >= TAG_End && tag <= TAG_IntArray)
	{
		const NBTData * data = read->GetData();
		if (!data)
			continue;
		if (!valid)
		{
			if (data->tag != TAG_Compound)
				continue;
			if (!data->name)
				continue;
			if (!StringCheck(data->name, "Level"))
				continue;
			valid = true;
		}
		switch (data->tag)
		{
			case TAG_Int:
				if (StringCheck(data->name, "xPos"))
				{
					chunkX = data->i;
				}
				else if (StringCheck(data->name, "zPos"))
				{
					chunkY = data->i;
				}
				break;
			case TAG_Long:
				if (StringCheck(data->name, "LastUpdate"))
				{
					lastUpdate = data->l;
				}
				break;
			case TAG_Byte:
				if(StringCheck(data->name, "Y") && section != 0)
				{
					section->y = data->i;
				}
				break;
			case TAG_IntArray:
				if(StringCheck(data->name, "HeightMap"))
				{
					if (data->size < MAPX * MAPY * sizeof(int))
						continue;
					for(int i = 0; i < MAPX * MAPY; i++)
					{
						anvilHeightMap[i] = Convert<int>(&(data->ias[i * sizeof(int)]));
					}
				}
				break;
			case TAG_List:
				if(StringCheck(data->name, "Sections"))
				{
					if(section != 0)
					{
						if(section->y >= 0 && section->y < MAXSECTIONS)
							this->sections[section->y] = section;
						else // shouldn't happen;
							delete section;
					}

					section = new Section();
					count++;
				}
				break;
			case TAG_ByteArray:
				// Blocks
				if (StringCheck(data->name, "Blocks") && section != 0)
				{
					if (data->size < SECTIONBLOCKS)
						continue;
					// Get blocks
					if (pref.logging)
					{
						for (int i = 0; i < SECTIONBLOCKS; ++i)
						{
							++amount.block[section->blocks[i] = data->bas[i]];
						}
					}
					else
					{
						memcpy(section->blocks, data->bas, SECTIONBLOCKS);
					}
					continue;
				}
				// Data
				if (StringCheck(data->name, "Data")  && section != 0)
				{
					if (data->size < SECTIONBLOCKS / 2)
						continue;

					memcpy(section->blockData, data->bas, SECTIONBLOCKS / 2);
					continue;
				}
				// Skylight
				if (StringCheck(data->name, "SkyLight")  && section != 0)
				{
					if (data->size < SECTIONBLOCKS / 2)
						continue;
					memcpy(section->skyLight, data->bas, SECTIONBLOCKS / 2);
					continue;
				}
				// Blocklight
				if (StringCheck(data->name, "BlockLight")  && section != 0)
				{
					if (data->size < SECTIONBLOCKS / 2)
						continue;
					memcpy(section->blockLight, data->bas, SECTIONBLOCKS / 2);
					continue;
				}
				break;
		}
	}

	if(section != 0)
	{
		if(section->y >= 0 && section->y < MAXSECTIONS)
			this->sections[section->y] = section;
		else // shouldn't happen;
			delete section;
	}

	// Invalid
	if (!valid)
	{
		amount.filesize = 0;
		return false;
	}
	
	// 10 000 000
	if (abs(chunkX) > MAPCOORD || abs(chunkY) > MAPCOORD)
	{
		amount.filesize = 0;
		valid = false;
		return false;
	}

	return true;
}

int AnvilChunk::GetHeight(int x, int y, int Z, int minZ)
{
	if (x < 0 || x >= MAPX) return 0;
	if (y < 0 || y >= MAPY) return 0;
	//if ((pref.flags & CHUNKP_NOHEIGHTMAP) == 0 && Z > (anvilHeightMap[y * MAPY + x]))
	//	Z = (anvilHeightMap[y * MAPY + x]);
	for (int z = Z; z >= minZ; --z)
		if (ReadBlock(x, y, z) != AIR || z <= minZ)
			return z;
	return minZ;
}

Color AnvilChunk::GetColor(int x, int y, int z, bool side)
{
	unsigned short n = ReadBlock(x, y, z);
	switch (n)
	{
		case WOOL:
			n = WOOLCOLOR + ReadBlockData(x, y, z);
			break;
		case LOG:
			n = LOGCOLOR + ReadBlockData(x, y, z) % 3;
			break;
		case LEAVES:
			n = LEAFCOLOR + ReadBlockData(x, y, z) % 3;
			break;
		case SLAB_STONE:
			n = SLABCOLOR + ReadBlockData(x, y, z) % 4;
			break;
		case GRASS:
			if (side)
				n = DIRT;
			break;
	}
	return DataValues::GetColor(n);
}

BYTE AnvilChunk::ReadBlock(int x, int z, int y)
{
	int section = y / 16;
	if(section< 0 || section >= MAXSECTIONS || section == 0 || sections[section] == 0)
		return 0;

	Transform(x, z);
	if(x < 0 || x > MAPX)
		return 0;
	if(z < 0 || z > MAPY)
		return 0;

	return sections[section]->blocks[((y % 16) * 16 + z) * 16 + x];
}

BYTE AnvilChunk::ReadBlockData(int x, int z, int y)
{
	int section = y / 16;
	if(section < 0 || section >= MAXSECTIONS || sections == 0 || sections[section] == 0)
		return 0;

	Transform(x, z);
	if(x < 0 || x > MAPX)
		return 0;
	if(z < 0 || z > MAPY)
		return 0;

	int i = ((y % 16) * 16 + z) * 16 + x;
	if(i % 2 == 0)
		return sections[section]->blockData[i/2] & 0x0F;
	else
		return (sections[section]->blockData[i/2] >> 4) & 0x0F;
}

BYTE AnvilChunk::ReadBlockLight(int x, int z, int y)
{
	int section = y / 16;
	if(section < 0 || section >= MAXSECTIONS || sections == 0 || sections[section] == 0)
		return 0;

	Transform(x, z);
	if(x < 0 || x > MAPX)
		return 0;
	if(z < 0 || z > MAPY)
		return 0;

	int i = ((y % 16) * 16 + z) * 16 + x;
	if(i % 2 == 0)
		return sections[section]->blockLight[i/2] & 0x0F;
	else
		return (sections[section]->blockLight[i/2] >> 4) & 0x0F;
}

void AnvilChunk::PassExclude(int & x, int & y, int & z, Color & color)
{
	if (z < 0)
		return;
	BYTE n = 0;
	do
	{
		n = ReadBlock(x, y, z);
	}
	while (!pref.block[n] && --z >= 0);
	color = (z < 0) ? 0 : GetColor(x, y, z);
}

#include <math.h>

void AnvilChunk::PassNight(int & x, int & y, int & z, Color & color)
{
	if (z < 0)
		return;
	BYTE bLight = ReadBlockLight(x, y, z);
	float light = pow(0.9f, 15 - bLight);
	color.Interpolate(Color(0, 0, 0), 1 - light);
}


//I can't figure out what I'm doing diffently, but
//without this chunks are rendered flipped over a
//diagonal
void AnvilChunk::Transform(int &x, int &z)
{
	//flip horizontal
	z = 15 - z;

	//rotate 270 degrees clockwise
	int temp = z;
	z = 15 - x;
	x = temp;

	temp = z;
	z = 15 - x;
	x = temp;

	temp = z;
	z = 15 - x;
	x = temp;
}