

#include "System.hpp"

#include <stdio.h>
#ifdef WIN32
#include <windows.h>
#include <direct.h>
#include <io.h>
#elif defined(UNIX)
#include <unistd.h>
#endif

// Get environement variable
STRING Port::GetEnv(const char * var)
{
	STRING env;
#ifdef WIN32
	size_t tLen = 0;
#ifdef UNICODE
	wchar_t path[FILENAME_MAX];
	wchar_t var2[256] = {0};
	size_t size = strlen(var)+1;
	mbstowcs_s(&tLen, var2, size, var, size);
	_wgetenv_s(&tLen, path, FILENAME_MAX, var2);
#else
	char path[FILENAME_MAX];
	getenv_s(&tLen, path, FILENAME_MAX, var);
#endif
	env = path;
#elif defined(UNIX)
	char * temp = getenv(var);
	env.assign(temp, temp+strlen(temp));
#endif
	return env;
}

// Get current working directory
STRING Port::GetCurrentDir()
{
	STRING dir;
#ifdef WIN32
	int tLen = 0;
#ifdef UNICODE
	wchar_t path[FILENAME_MAX] = {0};
	_wgetcwd(path, FILENAME_MAX);
#else
	char path[FILENAME_MAX] = {0};
	_getcwd(path, FILENAME_MAX);
#endif
	dir = path;
#elif defined(UNIX)
	char path[FILENAME_MAX] = {0};
	getcwd(path, FILENAME_MAX);
	dir.assign(path, path+strlen(path));
#endif
	return dir;
}

// Change current working directory
bool Port::SetCurrentDir(STRING str)
{
#ifdef WIN32
#ifdef UNICODE
	return _wchdir(str.c_str()) == 0;
#else
	return _chdir(str.c_str()) == 0;
#endif
#elif defined(UNIX)
	std::string s;
	s.assign(str.begin(), str.end());
	return chdir(s.c_str()) == 0;
#endif
}

// Replace a path to be correct
STRING Port::ReplacePath(STRING path)
{
#ifdef WIN32
#ifdef UNICODE
	path = Replace(path, L'/', L'\\');
#else
	path = Replace(path, '/', '\\');
#endif
#elif defined(UNIX)
#ifdef UNICODE
	path = Replace(path, L'\\', L'/');
#else
	path = Replace(path, '\\', '/');
#endif
#endif
	return path;
}

// Get file name from path
STRING Port::FileName(STRING file)
{
#ifdef WIN32
#ifdef UNICODE
	size_t pos = file.find_last_of(L"\\") + 1;
#else
	size_t pos = file.find_last_of("\\") + 1;
#endif
#elif defined(UNIX)
#ifdef UNICODE
	size_t pos = file.find_last_of(L"/") + 1;
#else
	size_t pos = file.find_last_of("/") + 1;
#endif
#endif
	return file.substr(pos);
}

// Get path from string
STRING Port::Directory(STRING str)
{
#ifdef WIN32
#ifdef UNICODE
	size_t pos = str.find_last_of(L"\\");
#else
	size_t pos = str.find_last_of("\\");
#endif
#elif defined(UNIX)
#ifdef UNICODE
	size_t pos = str.find_last_of(L"/");
#else
	size_t pos = str.find_last_of("/");
#endif
#endif
	return str.substr(0, pos);
}

// Path exists
bool Port::ExistPath(STRING path)
{
#ifdef WIN32
#ifdef UNICODE
	return _waccess_s(path.c_str(), 0) == 0;
#else
	return _access_s(path.c_str(), 0) == 0;
#endif
#elif defined(UNIX)
	std::string s;
	s.assign(path.begin(), path.end());
	return access(s.c_str(), 0) == 0;
#endif
}

// 1 - Exist
// 2 - File
// 4 - Dir with no ending slash
// 8 - Ending slash

// Tell how path is built up
int Port::TellPath(STRING path)
{
	// Nowhere
	if (path.empty())
		return 0;
	// If it exists
	int exist = 0;
	if (ExistPath(path))
		exist = 1;

	// Precalculate
#ifdef WIN32
#ifdef UNICODE
	int dirPos = (int)path.find_last_of(L"\\");
#else
	int dirPos = (int)path.find_last_of("\\");
#endif
#elif defined(UNIX)
#ifdef UNICODE
	int dirPos = (int)path.find_last_of(L"/");
#else
	int dirPos = (int)path.find_last_of("/");
#endif
#endif
#ifdef UNICODE
	int filePos = (int)path.find_last_of(L".");
#else
	int filePos = (int)path.find_last_of(".");
#endif

	// Is a file
	if (filePos > dirPos)
		return 2 | exist;

	// What kind of path
	int slash = 0;
	if (path.size() - 1 == dirPos)
		slash = 8;
	return 4 | slash | exist;
}

// Porting of sleep
void Port::Rest(UINT milli)
{
#ifdef WIN32
	Sleep(milli);
#elif defined(UNIX)
	usleep(milli * 1000);
#endif
}


#ifdef WIN32 // Windows
#include <windows.h>
#elif defined(__APPLE__)
#include "sys/types.h"
#include "sys/sysinfo.h"
#include "sys/sysctl.h"
#elif defined(UNIX)
#include <unistd.h>
// Compatibility with other systems
#ifdef _SC_PAGESIZE
#define SIZE_OF_PAGE _SC_PAGESIZE
#else
#define SIZE_OF_PAGE _SC_PAGE_SIZE
#endif
#endif // defined(UNIX)

// Get current available memory
size_t Port::GetAvailableMemory()
{
#ifdef WIN32
	MEMORYSTATUS info;
	GlobalMemoryStatus(&info);
	return info.dwAvailPhys;
#elif defined(__APPLE__)
	struct sysinfo info;
	sysinfo(&info);
	return total = ((info.freeram) + info.freeswap) * info.mem_unit;
#elif defined(UNIX)
	return sysconf(_SC_AVPHYS_PAGES) * sysconf(SIZE_OF_PAGE);
#endif
}

// Get total memory
size_t Port::GetTotalMemory()
{
#ifdef WIN32
	MEMORYSTATUS info;
	GlobalMemoryStatus(&info);
	return info.dwTotalPhys;
#elif defined(__APPLE__)
	struct sysinfo info;
	sysinfo(&info);
	return total = (info.totalram + info.totalswap) * info.mem_unit;
#elif defined(UNIX)
	return sysconf(_SC_PHYS_PAGES) * sysconf(SIZE_OF_PAGE);
#endif
}

// Get amount of CPU cores
size_t Port::GetLogicalCores()
{
#ifdef WIN32
	SYSTEM_INFO info;
	GetSystemInfo(&info);
	return info.dwNumberOfProcessors;
#elif defined(__APPLE__)
	int mib[4] = {0};
	mib[0] = CTL_HW;
	mib[1] = HW_AVAILCPU;
	size_t size = sizeof(unsigned long);
	unsigned long amount = 0;
	sysctl(mib, 2, &amount, &size, 0, 0);
	// Try again
	if (amount < 1)
	{
		mib[1] = HW_NCPU;
		sysctl(mib, 2, &amount, &size, 0, 0);
		// Last failure
		if (amount < 1)
			amount = 0;
	}
	return amount;
#elif defined(UNIX)
	return sysconf(_SC_NPROCESSORS_ONLN);
#endif
}