

#include "ThreadPool.hpp"

//#include "Utility.hpp"
#include "System.hpp"

#include "SafeQueue.hpp"

struct Thread
{
	Thread() : function(0), parent(0) {}

	THREAD_FUNC function;
	ThreadPool * parent;
	SafeQueue<WORK> orders;
	SafeQueue<RESULT> results;
};

// Each thread processing
THREAD_FUNCTION(ThreadPool_Work, data)
//DWORD WINAPI ThreadPool_Work(PVOID data)
{
	Thread * thread = (Thread *)data;

	ThreadPool * parent = thread->parent;

	WORK next = 0;
	RESULT result = 0;

	Port::Rest(1);

	while (parent->StayAlive())
	{
		result = 0;
		// Idle
		if (thread->orders.Empty())
		{
			Port::Rest(1);
			continue;
		}
		next = thread->orders.Pop();
		result = thread->function(next);

		// Wait for free space
		while (parent->StayAlive())
		{
			if (thread->results.Push(result))
			{
				break;
			}
			Port::Rest(1);
		}
	}

	parent->MarkThreadClosed();
	return 0;
}

// Reset everything
ThreadPool::ThreadPool()
: process(false), threads(0), closed(0), 
stacked(0)
{
	status.works = 0;
	status.results = 0;
}

// Safe quit
ThreadPool::~ThreadPool()
{
	if (process)
		Shutdown();
}

// Init
void ThreadPool::Initialize(int size, THREAD_FUNC func)
{
	// Idiotproof
	if (size <= 0)
		return;
	if (threads)
		return;

	process = true;
	threads = new Thread[size];
	for (int i = 0; i < size; ++i)
	{
		threads[i].parent = this;
		threads[i].function = func;
#ifdef WIN32
		group.push_back(CreateThread(NULL, 0, ThreadPool_Work, (void *)&threads[i], 0, 0));
#elif defined(UNIX)
		group.push_back(CreateThread(ThreadPool_Work, (void *)&threads[i]));
#endif
	}
}

// Shut everything down
void ThreadPool::Shutdown()
{
	process = false;

	HANDLE * temp = new HANDLE[group.size()];
	for (size_t i = 0; i < group.size(); ++i)
		temp[i] = group[i];

#ifdef WIN32
	WaitForMultipleObjects(group.size(), temp, TRUE, INFINITE);
#elif defined(UNIX)
	for (int i = 0; i < group.size(); ++i)
		WaitThread(group[i]);
#endif

	delete[] temp;

	if (threads)
	{
		delete[] threads;
		threads = NULL;
	}
}

// Push data onto the queues
bool ThreadPool::Work()
{
	// Fill working queues
	while (!FullWork() && !work.empty())
	{
		PushWork(work.front());
		work.pop();
	}
	// Take out results
	while (!EmptyResult())
	{
		RESULT r = 0;
		PopResult(&r);
		result.push(r);
	}

	// Put to sleep now
#ifdef WIN32
	SwitchToThread();
#else
	YieldThread();
#endif
	
	return stacked > 0;
}

// Add more work to the threads
bool ThreadPool::PushWork(WORK order)
{
	// Fill the most emptiest thread queue
	SafeQueue<WORK> * min = &threads[0].orders;
	int m = min->Amount();
	for (size_t i = 0; i < group.size(); ++i)
	{
		int n = threads[i].orders.Amount();
		if (n < m)
		{
			m = n;
			min = &threads[i].orders;
		}
	}
	bool t = min->Push(order);
	if (t)
	{
		++status.works;
	}
	else
		return false;
	return t;
}

// Check if the working queue is full
bool ThreadPool::FullWork()
{
	for (size_t i = 0; i < group.size(); ++i)
		if (!threads[i].orders.Full())
			return false;
	return true;
}

// Check if the working queue is empty
bool ThreadPool::EmptyWork()
{
	for (size_t i = 0; i < group.size(); ++i)
		if (!threads[i].orders.Empty())
			return false;
	return true;
}

// Check if the result queue is empty
bool ThreadPool::EmptyResult()
{
	for (size_t i = 0; i < group.size(); ++i)
		if (!threads[i].results.Empty())
			return false;
	return true;
}

// Get result from queue
bool ThreadPool::PopResult(RESULT * result)
{
	if (!result)
		return false;
	// Find most full queue
	SafeQueue<RESULT> * max = &threads[0].results;
	int m = max->Amount();
	for (size_t i = 0; i < group.size(); ++i)
	{
		int n = threads[i].results.Amount();
		if (n > m)
		{
			m = n;
			max = &threads[i].results;
		}
	}
	if (!max->Empty())
	{
		*result = max->Pop();
		++status.results;
		return true;
	}
	return false;
}

// Push to temporary storage
void ThreadPool::Push(WORK order)
{
	work.push(order);
	++stacked;
	Work();
}

// Get from results
bool ThreadPool::Pop(RESULT * r)
{
	if (!r)
		return false;
	Work();
	if (result.empty())
		return false;
	*r = result.front();
	result.pop();
	--stacked;
	return true;
}