

#include "Transform.hpp"

#include <math.h>
#include <complex>

using namespace std;

// Numerous of overloaded operators
const Vector3D Vector3D::operator+(const Vector3D u) const
{
	return Vector3D(x + u.x, y + u.y, z + u.z);
}

const Vector3D Vector3D::operator-(const Vector3D u) const
{
	return Vector3D(x - u.x, y - u.y, z - u.z);
}

Vector3D Vector3D::operator-() const
{
	return Vector3D(-x, -y, -z);
}

const Vector3D Vector3D::operator*(const double k) const
{
	return Vector3D(x * k, y * k, z * k);
}

const Vector3D Vector3D::operator/(const double k) const
{
	return Vector3D(x / k, y / k, z / k);
}

// Make a dot product
double Vector3D::DotProduct(Vector3D v, Vector3D u)
{
	return (v.x * u.x) + 
		(v.y * u.y) + 
		(v.z * u.z);
}

// Make a cross product
Vector3D Vector3D::CrossProduct(Vector3D v, Vector3D u)
{
	return Vector3D(
		(v.y * u.y) - (v.z * u.y), 
		(v.z * u.x) - (v.x * u.z), 
		(v.x * u.y) - (v.y * u.x));
}

// Get the angle from two vectors
double Vector3D::Angle(Vector3D v, Vector3D u)
{
	return (float)acos(Clamp(DotProduct(v, u) / (v.Length() * u.Length())));
}

// Get closest position
bool Vector3D::Compare(Vector3D v, Vector3D u)
{
	return Compare(v, u, 0.00001F);
}

bool Vector3D::Compare(Vector3D v, Vector3D u, double limit)
{
	return (abs(v.x - u.x) <= limit) &&
		(abs(v.y - u.y) <= limit) &&
		(abs(v.z - u.z) <= limit);
}

// Crop the float to between -1 and 1
double Vector3D::Clamp(double f)
{
	if (f > 1)
	{
		return 1;
	}

	if (f < -1)
	{
		return -1;
	}

	return f;
}

// Get the length of the vector
double Vector3D::Length()
{
	return abs(sqrt((x * x) + (y * y) + (z * z)));
}

// Normalizes a vector
Vector3D Vector3D::Normalize()
{
	return (*this) / Length();
}


// Matrix methods

// Constructor
Matrix4x4::Matrix4x4(double null)
{
	for (int y = 0; y < 4; ++y)
		for (int x = 0; x < 4; ++x)
			matrix[y][x] = 0;

	matrix[0][0] = null;
	matrix[1][1] = null;
	matrix[2][2] = null;
	matrix[3][3] = null;
}

// Translation
Matrix4x4 Matrix4x4::translate(double x, double y, double z)
{
	Matrix4x4 t(1);
	t(3, 0) = x;
	t(3, 1) = y;
	t(3, 2) = z;
	return t;
}

// Scaling
Matrix4x4 Matrix4x4::scale(double x, double y, double z)
{
	Matrix4x4 s(1);
	s(0, 0) = x;
	s(1, 1) = y;
	s(2, 2) = z;
	return s;
}

// Rotation on X
Matrix4x4 Matrix4x4::rotX(double rad)
{
	Matrix4x4 r(1);
	r(1, 1) = cos(rad);
	r(1, 2) = sin(rad);
	r(2, 1) = -sin(rad);
	r(2, 2) = cos(rad);
	return r;
}

// Rotation on Y
Matrix4x4 Matrix4x4::rotY(double rad)
{
	Matrix4x4 r(1);
	r(0, 0) = cos(rad);
	r(0, 2) = -sin(rad);
	r(2, 0) = sin(rad);
	r(2, 2) = cos(rad);
	return r;
}

// Rotation on Z
Matrix4x4 Matrix4x4::rotZ(double rad)
{
	Matrix4x4 r(1);
	r(0, 0) = cos(rad);
	r(0, 1) = sin(rad);
	r(1, 0) = -sin(rad);
	r(1, 1) = cos(rad);
	return r;
}

// Multiplication with Matrix4x4
Matrix4x4 Matrix4x4::operator*(Matrix4x4 b)
{
	Matrix4x4 c(0);
	for (int x = 0; x < 4; ++x)
		for (int y = 0; y < 4; ++y)
			for (int i = 0; i < 4; ++i)
				c.matrix[y][x] += matrix[i][x] * b.matrix[y][i];
	
	return c;
}

// Multiplication with Vector3D
Vector3D Matrix4x4::operator*(Vector3D a)
{
	Vector3D c(0,0,0);
	{
		c.x += a.x * matrix[0][0];
		c.x += a.y * matrix[0][1];
		c.x += a.z * matrix[0][2];
		c.x += matrix[3][0];

		c.y += a.x * matrix[1][0];
		c.y += a.y * matrix[1][1];
		c.y += a.z * matrix[1][2];
		c.y += matrix[3][1];

		c.z += a.x * matrix[2][0];
		c.z += a.y * matrix[2][1];
		c.z += a.z * matrix[2][2];
		c.z += matrix[3][2];
	}
	
	return c;
}

// Addition with Matrix4x4
Matrix4x4 Matrix4x4::operator+(Matrix4x4 b)
{
	Matrix4x4 c(0);
	for (int y = 0; y < 4; ++y)
		for (int x = 0; x < 4; ++x)
			c.matrix[y][x] = matrix[y][x] + b.matrix[y][x];
	
	return c;
}

// Rotation struct

#include "ChunkPrefs.hpp"

// Transform current image
void Rotation::Transform(double yaw, double pitch)
{
	yCos = cos(yaw);
	ySin = sin(yaw);
	pCos = cos(pitch);
	pSin = sin(pitch);
	int oldWidth = width;
	int oldHeight = height;
	// Change size
	{
		double _yCos = (yCos < 0) ? -yCos : yCos;
		double _ySin = (ySin < 0) ? -ySin : ySin;
		double _pCos = (pCos < 0) ? -pCos : pCos;
		double _pSin = (pSin < 0) ? -pSin : pSin;

		UINT newWidth = UINT((double)width * _yCos + (double)height * _ySin);
		UINT newHeight = UINT((double)newWidth * _pCos + (double)depth * _pSin);
		width = newWidth;
		height = newHeight;
	}
	int halfWidth = width / 2;
	int halfHeight = height / 2;
	int halfDepth = depth / 2;
	// Make Matrix
	//transform = MATRIX();
	transform = transform * 
		//transform.translate(-oldWidth / 2, -oldHeight / 2, -halfDepth) * 
		transform.rotX(pitch) * 
		transform.rotZ(-yaw) * 
		transform.translate(halfWidth, halfHeight, halfDepth);
	/*transform = transform * transform.translate(halfWidth, halfHeight, halfDepth);
	transform = transform * transform.rotZ(-yaw);
	transform = transform * transform.rotX(pitch);
	transform = transform * transform.translate(-oldWidth / 2, -oldHeight / 2, -halfDepth);*/
	/*transform = transform * transform.translate(oldWidth / 2, oldHeight / 2, halfDepth);
	transform = transform * transform.rotZ(-yaw);
	transform = transform * transform.rotX(pitch);
	transform = transform * transform.translate(-halfWidth, -halfHeight, -halfDepth);*/
}

// Place position in space with rotation
Vector3D Rotation::Place(double x, double y, double z)
{
	Vector3D d(x - ((double)width / 2), y - ((double)height / 2), z - ((double)depth / 2));
	//Vector3D d(x, y, z);
	d = transform * d;
	return d;
}

// Check if the image is rotated straight
bool Rotation::IsStraight()
{
#define FCOMP(a,b) (fabs((a)-(b))<0.0001f)
	return (FCOMP(ySin,0) || FCOMP(yCos,0)) && FCOMP(pSin,0);
}